/*
 * dev_ioc.c - IOC (Block Device Controler) generic device API implementation.
 *
 * Author  Alain Greiner    (2016)
 *
 * Copyright (c) UPMC Sorbonne Universites
 *
 * This file is part of ALMOS-MK
 *
 * ALMOS-MKH is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.0 of the License.
 *
 * ALMOS-MKH is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ALMOS-MKH; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <kernel_config.h>
#include <hal_types.h>
#include <hal_gpt.h>
#include <soclib_bdv.h>
#include <soclib_hba.h>
#include <thread.h>
#include <printk.h>
#include <chdev.h>
#include <dev_ioc.h>

/////////////////////////////////////////////////////////////////////////////////////////
// Extern global variables
/////////////////////////////////////////////////////////////////////////////////////////

extern chdev_directory_t  chdev_dir;     // allocated in kernel_init.c

//////////////////////////////////
void dev_ioc_init( chdev_t * ioc )
{
    // the PIC chdev must be initialized before the IOC chdev, because
    // the IOC chdev initialisation requires the routing of an external IRQ
    xptr_t  pic_xp  = chdev_dir.pic;

    assert( (pic_xp != XPTR_NULL) , __FUNCTION__ , "PIC not initialised before IOC" );

    // get implementation and channel from chdev descriptor
    uint32_t  impl    = ioc->impl;
    uint32_t  channel = ioc->channel;

    // set chdev name
    snprintf( ioc->name , 16 , "ioc_%d" , channel );

    // call driver init function
    if( impl == IMPL_IOC_BDV )
    {
        soclib_bdv_init( ioc );
    }
    else if( impl == IMPL_IOC_HBA )
    {
        soclib_hba_init( ioc );
    }
    else
    {
        assert( false , __FUNCTION__ , "undefined IOC device implementation" );
    }

    // select a core to execute the IOC server thread
    lid_t lid = cluster_select_local_core();

    // bind the IOC IRQ to the selected core
    dev_pic_bind_irq( lid , ioc );

    // enable IRQ
    dev_pic_enable_irq( lid ,ioc );

    // create server thread
    thread_t * new_thread;
    error_t    error;

    error = thread_kernel_create( &new_thread,
                                  THREAD_DEV,
                                  &chdev_sequencial_server,
                                  ioc,
                                  lid );

    assert( (error == 0) , __FUNCTION__ , "cannot create server thread" );

    // set "server" field in ioc descriptor
    ioc->server = new_thread;

    // start server thread
    thread_unblock( XPTR( local_cxy , new_thread ) , THREAD_BLOCKED_GLOBAL );

}  // end dev_ioc_init()

//////////////////////////////////////////////////////////////////////////////////
// This static function is called by dev_ioc_read() & dev_ioc_write() functions.
// It builds and registers the command in the calling thread descriptor.
// Then, it registers the calling thead in chdev waiting queue.
// Finally it blocks on the THREAD_BLOCKED_DEV condition and deschedule.
////////////////////////////////////i/////////////////////////////////////////////
static error_t dev_ioc_access( uint32_t   cmd_type,
                               uint8_t  * buffer,
                               uint32_t   lba,
                               uint32_t   count )
{
    thread_t * this = CURRENT_THREAD;              // pointer on client thread

    ioc_dmsg("\n[INFO] in %s : thread %x in process %x"
             " for lba = %x / buffer = %x / at cycle %d\n",
             __FUNCTION__ , this->trdid , this->process->pid ,
             lba , (intptr_t)buffer , hal_get_cycles() );

    // software L2/L3 cache coherence for memory buffer
    if( chdev_dir.iob )
    {
        if ( cmd_type == IOC_READ ) dev_mmc_inval( XPTR( local_cxy , buffer ) , count<<9 );
        else                        dev_mmc_sync ( XPTR( local_cxy , buffer ) , count<<9 );
    }

    // get extended pointer on IOC chdev descriptor
    xptr_t  dev_xp = chdev_dir.ioc[0];

    assert( (dev_xp != XPTR_NULL) , __FUNCTION__ , "undefined IOC chdev descriptor" );

    // register command in calling thread descriptor
    this->command.ioc.dev_xp    = dev_xp;
    this->command.ioc.type      = cmd_type;
    this->command.ioc.buf_xp    = XPTR( local_cxy , buffer );
    this->command.ioc.lba       = lba;
    this->command.ioc.count     = count;

    // register client thread in IOC chdev waiting queue, activate server thread,
    // block client thread on THREAD_BLOCKED_IO and deschedule.
    // it is re-activated by the ISR signaling IO operation completion.
    chdev_register_command( dev_xp , this );

    ioc_dmsg("\n[INFO] in %s : thread %x in process %x"
             " completes / error = %d / at cycle %d\n",
             __FUNCTION__ , this->trdid , this->process->pid ,
             this->command.ioc.error , hal_get_cycles() );

    // return I/O operation status
    return this->command.ioc.error;

}  // end dev_ioc_access()

////////////////////////////////////////////
error_t dev_ioc_read( uint8_t      * buffer,
                      uint32_t       lba,
                      uint32_t       count )
{
    return dev_ioc_access( IOC_READ , buffer , lba , count );
}

////////////////////////////////////////////
error_t dev_ioc_write( uint8_t     * buffer,
                       uint32_t      lba,
                       uint32_t      count )
{
    return dev_ioc_access( IOC_WRITE , buffer , lba , count );
}

/////////////////////////////////////////////
error_t dev_ioc_sync_read( uint8_t  * buffer,
                           uint32_t   lba,
                           uint32_t   count )
{
    // get pointer on calling thread
    thread_t * this = CURRENT_THREAD;

    // software L2/L3 cache coherence for memory buffer
    if( chdev_dir.iob ) dev_mmc_inval( XPTR( local_cxy , buffer ) , count<<9 );

    // get extended pointer on IOC[0] chdev
    xptr_t  dev_xp = chdev_dir.ioc[0];

    assert( (dev_xp != XPTR_NULL) , __FUNCTION__ , "undefined IOC chdev descriptor" );

    // register command in calling thread descriptor
    this->command.ioc.dev_xp    = dev_xp;
    this->command.ioc.type      = IOC_SYNC_READ;
    this->command.ioc.buf_xp    = XPTR( local_cxy , buffer );
    this->command.ioc.lba       = lba;
    this->command.ioc.count     = count;

    // get driver command function
    cxy_t       dev_cxy = GET_CXY( dev_xp );
    chdev_t   * dev_ptr = (chdev_t *)GET_PTR( dev_xp );
    dev_cmd_t * cmd = (dev_cmd_t *)hal_remote_lpt( XPTR( dev_cxy , &dev_ptr->cmd ) );

    // mask the IRQ
    thread_t * server = (thread_t *)hal_remote_lpt( XPTR( dev_cxy , &dev_ptr->server ) );
    core_t   * core = (core_t *)hal_remote_lpt( XPTR( dev_cxy , &server->core ) );
    lid_t      lid = (lid_t)hal_remote_lw( XPTR( dev_cxy , &core->lid ) );
    dev_pic_disable_irq( lid , dev_xp );

    // call directly driver command
    cmd( XPTR( local_cxy , this ) );

    // unmask the IRQ
    dev_pic_enable_irq( lid , dev_xp );

    // return I/O operation status from calling thread descriptor
    return this->command.ioc.error;
}

